#
# Tools for building OpenSlide and its dependencies
#
# Copyright (c) 2023 Benjamin Gilbert
# All rights reserved.
#
# This script is free software: you can redistribute it and/or modify it
# under the terms of the GNU Lesser General Public License, version 2.1,
# as published by the Free Software Foundation.
#
# This script is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
# for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this script. If not, see <http://www.gnu.org/licenses/>.
#

from __future__ import annotations

import argparse
from collections.abc import Iterable
from typing import Any


class TypedArgs:
    '''Simple argparse wrapper to give us type annotations on the parsed
    arguments.  The caller subclasses this class, defines fields on it, then
    configures and runs argparse.ArgumentParser via instance methods.  The
    methods cross-check the field types with the configuration given to
    argparse.  This ensures argparse doesn't get out of sync with the field
    types, without needing to implement automatic translation from field
    types to argparse (which could cause subtle problems if buggy).

    The implementation doesn't try to be comprehensive.  If you start using
    a new argparse feature, you'll probably have to extend this.'''

    def __init__(self, *args: Any, **kwargs: Any):
        self.parser = argparse.ArgumentParser(*args, **kwargs)
        self._unprocessed_fields = {k for k in self.__annotations__.keys()}

    def add_arg(
        self,
        *args: Any,
        parser: argparse.ArgumentParser | None = None,
        **kwargs: Any,
    ) -> None:
        # calculate field name and whether None is an allowed value
        expected_bool = kwargs.get('action') in ('store_true', 'store_false')
        for arg in args:
            if arg.startswith('--'):
                # option
                field: str = arg.lstrip('-').replace('-', '_')
                expected_none_forbidden = (
                    True if expected_bool else kwargs.get('required', False)
                )
                break
            elif not arg.startswith('-'):
                # positional parameter
                field = arg.replace('-', '_')
                expected_none_forbidden = kwargs.get('required', True)
                break
        else:
            raise ValueError(f'Option name not found in "{args}"')
        if 'dest' in kwargs:
            # field name overridden
            field = kwargs['dest']
        if kwargs.get('default') is not None:
            expected_none_forbidden = True

        # calculate expected field type
        typ = kwargs.get('type', str)
        if isinstance(typ, argparse.FileType):
            if 'b' in typ._mode:
                expected_type = 'BinaryIO'
            else:
                expected_type = 'TextIO'
        elif expected_bool:
            expected_type = 'bool'
        else:
            expected_type = typ.__name__
        nargs: str | int | None = kwargs.get('nargs')
        if nargs in {'*', '+'} or type(nargs) is int:
            expected_type = f'list[{expected_type}]'

        # calculate actual field type and requiredness
        annotation = self.__annotations__.get(field)
        if annotation is None:
            raise AttributeError(f'TypedArgs has no field "{field}"')
        anno_types = {a.strip() for a in annotation.split('|')}
        try:
            # "| None" -> optional field
            anno_types.remove('None')
            anno_none_forbidden = False
        except KeyError:
            anno_none_forbidden = True
        if len(anno_types) != 1:
            # field allows only None, or 2+ things besides None
            raise AttributeError(f'Could not parse types for field "{field}"')
        anno_type = anno_types.pop()

        # do checks
        if anno_type != expected_type:
            raise TypeError(
                f'TypedArgs field "{field}" type "{anno_type}" != "{expected_type}"'  # noqa: E501
            )
        if anno_none_forbidden != expected_none_forbidden:
            raise TypeError(
                f'TypedArgs field "{field}" forbids None "{anno_none_forbidden}" != "{expected_none_forbidden}"'  # noqa: E501
            )

        # mark the field processed.  allow a field to be processed multiple
        # times on different subparsers.
        self._unprocessed_fields.discard(field)

        # set up the actual argument
        (parser or self.parser).add_argument(*args, **kwargs)

    def parse(self, allow_extra_fields: Iterable[str] | None = None) -> None:
        unprocessed = self._unprocessed_fields - set(allow_extra_fields or [])
        if unprocessed:
            raise ValueError(f'Unused fields: {unprocessed}')
        self.parser.parse_args(namespace=self)



def gen_range_tables(out, name, s_suffix, l_suffix, ranges):
    numshort = 0
    numlong = 0
    sptr = "NULL"
    lptr = "NULL"

    for range in ranges:
        (low, high) = range
        if high < 0x10000:
            if numshort == 0:
                sptr = name + s_suffix
                pline = "static const xmlChSRange %s[] = {" % sptr
            else:
                pline += ","
            numshort += 1
        else:
            if numlong == 0:
                if numshort > 0:
                    out.write(pline + "};\n")
                lptr = name + l_suffix
                pline = "static const xmlChLRange %s[] = {" % lptr
            else:
                pline += ","
            numlong += 1
        if len(pline) > 60:
            out.write(pline + "\n")
            pline = "    "
        elif pline[-1:] == ",":
            pline += " "
        pline += "{%s, %s}" % (hex(low), hex(high))

    out.write(pline + "};\n")

    return "{%s,%s,%s,%s}" % (numshort, numlong, sptr, lptr)



#!/usr/bin/env python3
#
# Tools for building OpenSlide and its dependencies
#
# Copyright (c) 2023 Benjamin Gilbert
# All rights reserved.
#
# This script is free software: you can redistribute it and/or modify it
# under the terms of the GNU Lesser General Public License, version 2.1,
# as published by the Free Software Foundation.
#
# This script is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
# for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this script. If not, see <http://www.gnu.org/licenses/>.
#

from __future__ import annotations

from pathlib import Path
import shutil

from common.argparse import TypedArgs


class Args(TypedArgs):
    dll: Path
    output: Path


args = Args('write-import-library', description='Write import library.')
args.add_arg('dll', type=Path, help='built DLL')
args.add_arg('output', type=Path, help='output import library')
args.parse()

# We don't actually generate an import library, we just copy the one that's
# already built, changing the file extension in the process.  This is
# necessary because the import library is generated as a side effect of
# building the DLL and isn't addressable in Meson.  By adding an explicit
# build step, we make it addressable.
basename = '-'.join(args.dll.name.split('-')[0:-1])
shutil.copy2(args.dll.with_name(f'{basename}.dll.a'), args.output)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Copyright © 2025 Marco Trevisan <mail@3v1n0.net>
#
# SPDX-License-Identifier: LGPL-2.1-or-later
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
# MA  02110-1301  USA

""" Integration tests for gi-inspect-typelib. """

import os
import sys
import unittest

import taptestrunner
import testprogramrunner


class TestGIInspectTypelibBase(testprogramrunner.TestProgramRunner):
    """Integration test base class for checking gi-inspect-typelib behavior"""

    PROGRAM_NAME = "gi-inspect-typelib"
    PROGRAM_TYPE = testprogramrunner.ProgramType.NATIVE

    def runTestProgram(self, *args, **kwargs):
        return super().runTestProgram(args, **kwargs)


class TestGIInspectTypelibCommandLine(TestGIInspectTypelibBase):
    def test_help(self):
        """Test the --help argument."""
        result = self.runTestProgram("--help")
        self.assertIn("Usage:", result.out)
        self.assertIn(
            f"{self._program_name} [OPTION…] NAMESPACE - Inspect GI typelib", result.out
        )
        self.assertIn("--typelib-version=VERSION", result.out)
        self.assertIn("--print-shlibs", result.out)
        self.assertIn("--print-typelibs", result.out)

    def test_no_args(self):
        """Test running with no arguments at all."""
        result = self.runTestProgram(should_fail=True)
        self.assertEqual("Please specify exactly one namespace", result.err)

    def test_invalid_typelib(self):
        res = self.runTestProgram(
            "--print-typelibs", "--print-shlibs", "AnInvalidNameSpace", should_fail=True
        )
        self.assertFalse(res.out)
        self.assertIn(
            "Typelib file for namespace 'AnInvalidNameSpace' (any version) not found",
            res.err,
        )


class TestGIInspectTypelibForGLibTypelib(TestGIInspectTypelibBase):
    """Test introspection of typelib for GLib typelib"""

    TYPELIB_NAMESPACE = "GLib"
    TYPELIB_VERSION = "2.0"
    LIB_SONAME = "0"

    @classmethod
    def setUpClass(cls):
        super().setUpClass()

        if "G_TEST_BUILDDIR" in os.environ:
            os.environ["GI_TYPELIB_PATH"] = os.path.join(
                os.environ["G_TEST_BUILDDIR"], "..", "introspection"
            )

    def runTestProgram(self, *args, **kwargs):
        argv = list(args)
        argv.append(self.TYPELIB_NAMESPACE)

        if self.TYPELIB_VERSION:
            argv.append(f"--typelib-version={self.TYPELIB_VERSION}")

        return super().runTestProgram(*argv, **kwargs)

    def get_shlib_ext(self):
        if os.name == "nt":
            return f"-{self.LIB_SONAME}.dll"
        elif sys.platform == "darwin":
            return f".{self.LIB_SONAME}.dylib"

        return f".so.{self.LIB_SONAME}"

    def check_shlib(self, out):
        self.assertIn(
            f"lib{self.TYPELIB_NAMESPACE.lower()}-{self.TYPELIB_VERSION}"
            + f"{self.get_shlib_ext()}",
            out,
        )

    def check_typelib_deps(self, out):
        self.assertNotIn("GLib-2.0", out)
        self.assertNotIn("GModule-2.0", out)
        self.assertNotIn("GObject-2.0", out)
        self.assertNotIn("Gio-2.0", out)

    def test_print_typelibs(self):
        res = self.runTestProgram("--print-typelibs")
        self.assertFalse(res.err)
        if self.TYPELIB_NAMESPACE == "GLib":
            self.assertFalse(res.out)
        self.check_typelib_deps(res.out)

    def test_print_shlibs(self):
        res = self.runTestProgram("--print-shlibs")
        self.assertFalse(res.err)
        self.check_shlib(res.out)
        self.assertNotIn("GLib-2.0", res.out)

    def test_print_typelibs_and_shlibs(self):
        res = self.runTestProgram("--print-typelibs", "--print-shlibs")
        self.assertFalse(res.err)
        self.check_shlib(res.out)


class TestGIInspectTypelibForGObjectTypelib(TestGIInspectTypelibForGLibTypelib):
    """Test introspection of typelib for GObject typelib"""

    TYPELIB_NAMESPACE = "GObject"

    def check_typelib_deps(self, out):
        self.assertIn("GLib-2.0", out)
        self.assertNotIn("GModule-2.0", out)
        self.assertNotIn("GObject-2.0", out)
        self.assertNotIn("Gio-2.0", out)


class TestGIInspectTypelibForGioTypelib(TestGIInspectTypelibForGLibTypelib):
    """Test introspection of typelib for Gio typelib"""

    TYPELIB_NAMESPACE = "Gio"

    def check_typelib_deps(self, out):
        self.assertIn("GLib-2.0", out)
        self.assertIn("GObject-2.0", out)
        self.assertIn("GModule-2.0", out)
        self.assertNotIn("Gio-2.0", out)


if __name__ == "__main__":
    unittest.main(testRunner=taptestrunner.TAPTestRunner())



#!/usr/bin/env python3
#
# Original script modified in November 2003 to take advantage of
# the character-validation range routines, and updated to the
# current Unicode information (Version 4.0.1)
#
# NOTE: there is an 'alias' facility for blocks which are not present in
#	the current release, but are needed for ABI compatibility.  This
#	must be accomplished MANUALLY!  Please see the comments below under
#     'blockAliases'
#
import sys
import string
import rangetab

#
# blockAliases is a small hack - it is used for mapping block names which
# were were used in the 3.1 release, but are missing or changed in the current
# release.  The format is "OldBlockName:NewBlockName1[,NewBlockName2[,...]]"
blockAliases = []
blockAliases.append("CombiningMarksforSymbols:CombiningDiacriticalMarksforSymbols")
blockAliases.append("Greek:GreekandCoptic")
blockAliases.append("PrivateUse:PrivateUseArea,SupplementaryPrivateUseArea-A," + 
	"SupplementaryPrivateUseArea-B")

# minTableSize gives the minimum number of ranges which must be present
# before a range table is produced.  If there are less than this
# number, inline comparisons are generated
minTableSize = 8

blockfile = "Blocks-4.0.1.txt"
catfile = "UnicodeData-4.0.1.txt"


#
# Now process the "blocks" file, reducing it to a dictionary
# indexed by blockname, containing a tuple with the applicable
# block range
#
BlockNames = {}
try:
    blocks = open(blockfile, "r")
except:
    print("Missing %s, aborting ..." % blockfile)
    sys.exit(1)

for line in blocks.readlines():
    if line[0] == '#':
        continue
    line = line.strip()
    if line == '':
        continue
    try:
        fields = line.split(';')
        range = fields[0].strip()
        (start, end) = range.split("..")
        name = fields[1].strip()
        name = name.replace(' ', '')
    except:
        print("Failed to process line: %s" % (line))
        continue
    start = int(start, 16)
    end = int(end, 16)
    try:
        BlockNames[name].append((start, end))
    except:
        BlockNames[name] = [(start, end)]
blocks.close()
print("Parsed %d blocks descriptions" % (len(BlockNames.keys())))

for block in blockAliases:
    alias = block.split(':')
    alist = alias[1].split(',')
    for comp in alist:
        if comp in BlockNames:
            if alias[0] not in BlockNames:
                BlockNames[alias[0]] = []
            for r in BlockNames[comp]:
                BlockNames[alias[0]].append(r)
        else:
            print("Alias %s: %s not in Blocks" % (alias[0], comp))
            continue

#
# Next process the Categories file. This is more complex, since
# the file is in code sequence, and we need to invert it.  We use
# a dictionary with index category-name, with each entry containing
# all the ranges (codepoints) of that category.  Note that category
# names comprise two parts - the general category, and the "subclass"
# within that category.  Therefore, both "general category" (which is
# the first character of the 2-character category-name) and the full
# (2-character) name are entered into this dictionary.
#
try:
    data = open(catfile, "r")
except:
    print("Missing %s, aborting ..." % catfile)
    sys.exit(1)

nbchar = 0;
Categories = {}
for line in data.readlines():
    if line[0] == '#':
        continue
    line = line.strip()
    if line == '':
        continue
    try:
        fields = line.split(';')
        point = fields[0].strip()
        value = 0
        while point != '':
            value = value * 16
            if point[0] >= '0' and point[0] <= '9':
                value = value + ord(point[0]) - ord('0')
            elif point[0] >= 'A' and point[0] <= 'F':
                value = value + 10 + ord(point[0]) - ord('A')
            elif point[0] >= 'a' and point[0] <= 'f':
                value = value + 10 + ord(point[0]) - ord('a')
            point = point[1:]
        name = fields[2]
    except:
        print("Failed to process line: %s" % (line))
        continue
    
    nbchar = nbchar + 1
    # update entry for "full name"
    try:
        Categories[name].append(value)
    except:
        try:
            Categories[name] = [value]
        except:
            print("Failed to process line: %s" % (line))
    # update "general category" name
    try:
        Categories[name[0]].append(value)
    except:
        try:
            Categories[name[0]] = [value]
        except:
            print("Failed to process line: %s" % (line))

data.close()
print("Parsed %d char generating %d categories" % (nbchar, len(Categories.keys())))

#
# The data is now all read.  Time to process it into a more useful form.
#
# reduce the number list into ranges
for cat in Categories.keys():
    list = Categories[cat]
    start = -1
    prev = -1
    end = -1
    ranges = []
    for val in list:
        if start == -1:
            start = val
            prev = val
            continue
        elif val == prev + 1:
            prev = val
            continue
        elif prev == start:
            ranges.append((prev, prev))
            start = val
            prev = val
            continue
        else:
            ranges.append((start, prev))
            start = val
            prev = val
            continue
    if prev == start:
        ranges.append((prev, prev))
    else:
        ranges.append((start, prev))
    Categories[cat] = ranges

#
# Assure all data is in alphabetic order, since we will be doing binary
# searches on the tables.
#
bkeys = sorted(BlockNames.keys())

ckeys = sorted(Categories.keys())

#
# Generate the resulting files
#
try:
    output = open("codegen/unicode.inc", "w")
except:
    print("Failed to open codegen/unicode.inc")
    sys.exit(1)

#
# For any categories with more than minTableSize ranges we generate
# a range table suitable for xmlCharInRange
#
for name in ckeys:
    if len(Categories[name]) <= minTableSize or name == 'Cs':
        continue
    ranges = Categories[name]
    group = rangetab.gen_range_tables(output, 'xml' + name, 'S', 'L', ranges)
    output.write("static const xmlChRangeGroup xml%sG = %s;\n\n" %
                 (name, group))

for name in ckeys:
    if name == 'Cs':
        continue
    ranges = Categories[name]
    output.write("static int\nxmlUCSIsCat%s(int code) {\n" % name)
    if len(Categories[name]) > minTableSize:
        output.write("    return(xmlCharInRange((unsigned int)code, &xml%sG)"
            % name)
    else:
        start = 1
        for range in ranges:
            (begin, end) = range;
            if start:
                output.write("    return(");
                start = 0
            else:
                output.write(" ||\n           ");
            if (begin == end):
                output.write("(code == %s)" % (hex(begin)))
            else:
                output.write("((code >= %s) && (code <= %s))" % (
                         hex(begin), hex(end)))
    output.write(");\n}\n\n")

#
# Range tables for blocks
#

blockGroups = ''
for block in bkeys:
    name = block.replace('-', '')
    ranges = BlockNames[block]
    group = rangetab.gen_range_tables(output, 'xml' + name, 'S', 'L', ranges)
    output.write("\n")
    if blockGroups != '':
        blockGroups += ",\n"
    blockGroups += '  {"%s",\n   %s}' % (block, group)

output.write("static const xmlUnicodeRange xmlUnicodeBlocks[] = {\n")
output.write(blockGroups)
output.write("\n};\n\n")

output.close()



#!/usr/bin/env python3

import os
import sys
import re
import subprocess

# Put output files in the builddir
os.chdir(os.environ['MESON_BUILD_ROOT'])

infile = 'ro-eh-frame-conftest.c'
outfile = 'ro-eh-frame-conftest.o'

with open(infile, 'w') as f:
    f.write('extern void foo (void); void bar (void) { foo (); foo (); }')

args = sys.argv[1:]
args += ['-c', infile, '-o', outfile]
res = subprocess.run(args, stdout=subprocess.PIPE, universal_newlines=True, check=True)

stdout = subprocess.check_output(['readelf', '-WS', outfile], stdout=subprocess.PIPE, universal_newlines=True)
if re.search('eh_frame .* WA', stdout):
    sys.exit(0)

sys.exit(1)



#!/usr/bin/env python3
#
# Portions of this script have been (shamelessly) stolen from the
# prior work of Daniel Veillard (genUnicode.py)
#
# I, however, take full credit for any bugs, errors or difficulties :-)
#
# William Brack
# October 2003
#
# 18 October 2003
# Modified to maintain binary compatibility with previous library versions
# by adding a suffix 'Q' ('quick') to the macro generated for the original,
# function, and adding generation of a function (with the original name) which
# instantiates the macro.
#

import sys
import rangetab

#
# A routine to take a list of yes/no (1, 0) values and turn it
# into a list of ranges.  This will later be used to determine whether
# to generate single-byte lookup tables, or inline comparisons
#
def makeRange(lst):
    ret = []
    pos = 0
    while pos < len(lst):
        try:            # index generates exception if not present
            s = lst[pos:].index(1)      # look for start of next range
        except:
            break                       # if no more, finished
        pos += s                        # pointer to start of possible range
        try:
            e = lst[pos:].index(0)      # look for end of range
            e += pos
        except:                         # if no end, set to end of list
            e = len(lst)
        ret.append((pos, e-1))          # append range tuple to list
        pos = e + 1                     # ready to check for next range
    return ret

# minTableSize gives the minimum number of ranges which must be present
# before a 256-byte lookup table is produced.  If there are less than this
# number, a macro with inline comparisons is generated
minTableSize = 6

# dictionary of functions, key=name, element contains char-map and range-list
Functs = {}

state = 0

try:
    defines = open("codegen/ranges.def", "r")
except:
    print("Missing codegen/ranges.def, aborting ...")
    sys.exit(1)

#
# The lines in the .def file have three types:-
#   name:   Defines a new function block
#   ur:     Defines individual or ranges of unicode values
#   end:    Indicates the end of the function block
#
# These lines are processed below.
#
for line in defines.readlines():
    # ignore blank lines, or lines beginning with '#'
    if line[0] == '#':
        continue
    line = line.strip()
    if line == '':
        continue
    # split line into space-separated fields, then split on type
    try:
        fields = line.split(' ')
        #
        # name line:
        #   validate any previous function block already ended
        #   validate this function not already defined
        #   initialize an entry in the function dicitonary
        #       including a mask table with no values yet defined
        #
        if fields[0] == 'name':
            name = fields[1]
            if state != 0:
                print("'name' %s found before previous name" \
                      "completed" % (fields[1]))
                continue
            state = 1
            if name in Functs:
                print("name '%s' already present - may give" \
                      " wrong results" % (name))
            else:
                # dict entry with two list elements (chdata, rangedata)
                Functs[name] = [ [], [] ]
                for v in range(256):
                    Functs[name][0].append(0)
        #
        # end line:
        #   validate there was a preceding function name line
        #   set state to show no current function active
        #
        elif fields[0] == 'end':
            if state == 0:
                print("'end' found outside of function block")
                continue
            state = 0

        #
        # ur line:
        #   validate function has been defined
        #   process remaining fields on the line, which may be either
        #       individual unicode values or ranges of values
        #
        elif fields[0] == 'ur':
            if state != 1:
                raise Exception("'ur' found outside of 'name' block")
            for el in fields[1:]:
                pos = el.find('..')
                # pos <=0 means not a range, so must be individual value
                if pos <= 0:
                    # cheap handling of hex or decimal values
                    if el[0:2] == '0x':
                        value = int(el[2:],16)
                    elif el[0] == "'":
                        value = ord(el[1])
                    else:
                        value = int(el)
                    if ((value < 0) | (value > 0x1fffff)):
                        raise Exception('Illegal value (%s) in ch for'\
                                ' name %s' % (el,name))
                    # for ur we have only ranges (makes things simpler),
                    # so convert val to range
                    currange = (value, value)
                # pos > 0 means this is a range, so isolate/validate
                # the interval
                else:
                    # split the range into it's first-val, last-val
                    (first, last) = el.split("..")
                    # convert values from text into binary
                    if first[0:2] == '0x':
                        start = int(first[2:],16)
                    elif first[0] == "'":
                        start = ord(first[1])
                    else:
                        start = int(first)
                    if last[0:2] == '0x':
                        end = int(last[2:],16)
                    elif last[0] == "'":
                        end = ord(last[1])
                    else:
                        end = int(last)
                    if (start < 0) | (end > 0x1fffff) | (start > end):
                        raise Exception("Invalid range '%s'" % el)
                    currange = (start, end)
                # common path - 'currange' has the range, now take care of it
                # We split on single-byte values vs. multibyte
                if currange[1] < 0x100: # single-byte
                    for ch in range(currange[0],currange[1]+1):
                        # validate that value not previously defined
                        if Functs[name][0][ch]:
                            msg = "Duplicate ch value '%s' for name '%s'" % (el, name)
                            raise Exception(msg)
                        Functs[name][0][ch] = 1
                else:                   # multi-byte
                    if currange in Functs[name][1]:
                        raise Exception("range already defined in" \
                                " function")
                    else:
                        Functs[name][1].append(currange)

    except:
        print("Failed to process line: %s" % (line))
        raise

try:
    output = open("codegen/ranges.inc", "w")
except:
    print("Failed to open codegen/ranges.inc")
    sys.exit(1)

#
# Now output the generated data.
#

fkeys = sorted(Functs.keys())

for f in fkeys:

# First we convert the specified single-byte values into a group of ranges.
    if max(Functs[f][0]) > 0:   # only check if at least one entry
        rangeTable = makeRange(Functs[f][0])
        numRanges = len(rangeTable)
        if numRanges >= minTableSize:   # table is worthwhile
            # write the constant data to the code file
            output.write("const unsigned char %s_tab[256] = {\n" % f)
            pline = "   "
            for n in range(255):
                pline += " 0x%02x," % Functs[f][0][n]
                if len(pline) > 72:
                    output.write(pline + "\n")
                    pline = "   "
            output.write(pline + " 0x%02x };\n\n" % Functs[f][0][255])

#
# Next we do the unicode ranges
#

for f in fkeys:
    if len(Functs[f][1]) > 0:   # only generate if unicode ranges present
        rangeTable = Functs[f][1]
        rangeTable.sort()       # ascending tuple sequence
        group = rangetab.gen_range_tables(output, f, '_srng', '_lrng',
                                          rangeTable)

        output.write("const xmlChRangeGroup %sGroup =\n\t%s;\n\n" %
                     (f, group))

output.close()




#!/usr/bin/env python3
#
# Tools for building OpenSlide and its dependencies
#
# Copyright (c) 2023 Benjamin Gilbert
# All rights reserved.
#
# This script is free software: you can redistribute it and/or modify it
# under the terms of the GNU Lesser General Public License, version 2.1,
# as published by the Free Software Foundation.
#
# This script is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
# for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this script. If not, see <http://www.gnu.org/licenses/>.
#

from __future__ import annotations

import argparse
from contextlib import ExitStack
from pathlib import Path
import tempfile
from typing import BinaryIO, cast

from common.archive import (
    DirMember,
    FileMember,
    SymlinkMember,
    TarArchiveReader,
    TarArchiveWriter,
)
from common.argparse import TypedArgs
from common.macos import all_equal, merge_macho

DSYM_ARCHES = {'aarch64', 'x86_64'}


class Args(TypedArgs):
    bdists: list[BinaryIO]
    output: BinaryIO


args = Args(
    'write-universal-bdist', description='Write macOS universal bdist archive.'
)
args.add_arg(
    '-o',
    '--output',
    type=argparse.FileType('wb'),
    required=True,
    help='output file',
)
args.add_arg(
    'bdists',
    metavar='bdist',
    nargs='+',
    type=argparse.FileType('rb'),
    help='input file',
)
args.parse()

with ExitStack() as stack:
    tempdir = Path(
        stack.enter_context(
            tempfile.TemporaryDirectory(prefix='openslide-bin-')
        )
    )
    readers = stack.enter_context(TarArchiveReader.group(args.bdists))
    out = stack.enter_context(TarArchiveWriter(args.output))
    for members in readers:
        if all_equal(type(m) for m in members):
            all_type: type | None = type(members[0])
        else:
            all_type = None
        if not all_equal(members.relpaths):
            # path mismatch, which we only allow for dSYM relocations.
            # ensure we have a path component which is a dSYM arch
            if not all(
                DSYM_ARCHES.intersection(p.parts) for p in members.relpaths
            ):
                raise Exception(f'Path mismatch: {members.relpaths}')
            if all_type in (DirMember, FileMember):
                for member in members:
                    out.add(member.with_base(out.base))
            else:
                raise Exception(
                    'Unknown/mismatched types for relocations: '
                    f'{members.relpaths}'
                )
        elif all_type is DirMember or (
            all_type is SymlinkMember
            and all_equal(cast(SymlinkMember, m).target for m in members)
        ):
            out.add(members[0].with_base(out.base))
        elif all_type is FileMember:
            if (
                len(members.datas[0]) >= 4
                and members.datas[0][0:4] == b'\xcf\xfa\xed\xfe'
            ):
                macho_path = merge_macho(
                    [Path(cast(FileMember, m).fh.name) for m in members],
                    tempdir,
                )
                out.add(
                    FileMember(
                        out.base / members[0].relpath,
                        open(macho_path, 'rb'),
                    )
                )
            elif all_equal(members.datas):
                out.add(members[0].with_base(out.base))
            else:
                raise Exception(f'Contents mismatch: {members.relpaths}')
        else:
            raise Exception(f'Unknown/mismatched types: {members.relpaths}')



#
# Tools for building OpenSlide and its dependencies
#
# Copyright (c) 2023 Benjamin Gilbert
# All rights reserved.
#
# This script is free software: you can redistribute it and/or modify it
# under the terms of the GNU Lesser General Public License, version 2.1,
# as published by the Free Software Foundation.
#
# This script is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
# for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this script. If not, see <http://www.gnu.org/licenses/>.
#

from __future__ import annotations

from collections.abc import Iterable, Sequence
import os
from pathlib import Path
import subprocess
from typing import Any


def merge_macho(paths: Sequence[Path], outdir: Path) -> Path:
    outpath = outdir / paths[0].name
    args: list[str | Path] = [
        os.environ.get('LIPO', 'lipo'),
        '-create',
        '-output',
        outpath,
    ]
    args.extend(paths)
    subprocess.check_call(args)
    return outpath


def all_equal(items: Iterable[Any]) -> bool:
    it = iter(items)
    first = next(it)
    return all(i == first for i in it)



#!/usr/bin/env python3
#
# Tools for building OpenSlide and its dependencies
#
# Copyright (c) 2023-2025 Benjamin Gilbert
# All rights reserved.
#
# This script is free software: you can redistribute it and/or modify it
# under the terms of the GNU Lesser General Public License, version 2.1,
# as published by the Free Software Foundation.
#
# This script is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
# for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this script. If not, see <http://www.gnu.org/licenses/>.
#

from __future__ import annotations

import argparse
from typing import TextIO

from common.argparse import TypedArgs
from common.python import pyproject_fill_template


class Args(TypedArgs):
    input: TextIO
    output: TextIO


args = Args('write-pyproject', description='Write pyproject.toml.')
args.add_arg(
    'input',
    type=argparse.FileType('r'),
    help='input template file',
)
args.add_arg(
    'output',
    type=argparse.FileType('w'),
    help='output file',
)
args.parse()

with args.input:
    pyproject = pyproject_fill_template(args.input.read())
with args.output:
    args.output.write(pyproject)



#!/usr/bin/env python3

import os
import sys
import subprocess

# Put output files in the builddir
os.chdir(os.environ['MESON_BUILD_ROOT'])

infile = 'zarch-conftest.c'
outfile = 'zarch-conftest.S'

with open(infile, 'w') as f:
    f.write('void foo(void) { bar(); bar(); }')

args = sys.argv[1:]
args += ['-S', infile, '-o', outfile]
res = subprocess.run(args, stdout=subprocess.PIPE, universal_newlines=True, check=True)

with open(outfile, 'r') as f:
    if 'brasl' in f.read():
        sys.exit(0)

sys.exit(1)



#!/usr/bin/env python3
#
# Tools for building OpenSlide and its dependencies
#
# Copyright (c) 2023 Benjamin Gilbert
# All rights reserved.
#
# This script is free software: you can redistribute it and/or modify it
# under the terms of the GNU Lesser General Public License, version 2.1,
# as published by the Free Software Foundation.
#
# This script is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
# for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this script. If not, see <http://www.gnu.org/licenses/>.
#

from __future__ import annotations

import argparse
from pathlib import Path, PurePath
import re
from typing import BinaryIO

from common.archive import (
    ArchiveWriter,
    FileMember,
    SymlinkMember,
    TarArchiveWriter,
    ZipArchiveWriter,
)
from common.argparse import TypedArgs
from common.meson import meson_host
from common.software import Project


class Args(TypedArgs):
    artifacts: list[Path]
    output: BinaryIO


args = Args('write-bdist', description='Write bdist archive.')
args.add_arg(
    '-o',
    '--output',
    type=argparse.FileType('wb'),
    required=True,
    help='output file',
)
args.add_arg(
    'artifacts',
    metavar='artifact',
    nargs='+',
    type=Path,
    help='built artifact',
)
args.parse()

if meson_host() == 'windows':
    arc: ArchiveWriter = ZipArchiveWriter(args.output)
else:
    arc = TarArchiveWriter(args.output)
with arc:
    for path in args.artifacts:
        name = path.name
        if re.search('\\.(lib|dylib(\\.dSYM)?|so[.0-9]*(\\.debug)?)$', name):
            arcdir = arc.base / 'lib'
        elif name.endswith('.h'):
            arcdir = arc.base / 'include' / 'openslide'
        elif name in (
            'CHANGELOG.md',
            'VERSIONS.md',
            'versions.json',
            'licenses',
        ):
            arcdir = arc.base
        else:
            arcdir = arc.base / 'bin'

        if path.is_dir():
            arc.add_tree(arcdir, path)
        else:
            arc.add(FileMember(arcdir / name, path.open('rb')))
            if re.search('\\.so(\\.[0-9]+){3}$', name):
                for pat in '(\\.[0-9]+){2}$', '(\\.[0-9]+)+$':
                    lname = re.sub(pat, '', name)
                    arc.add(SymlinkMember(arcdir / lname, PurePath(name)))
            elif re.search('\\.[0-9]+\\.dylib$', name):
                lname = re.sub('\\.[0-9]+\\.dylib$', '.dylib', name)
                arc.add(SymlinkMember(arcdir / lname, PurePath(name)))

    # special case: copy OpenSlide README to root
    arc.add(
        FileMember(
            arc.base / 'README.md',
            open(Project.get('openslide').source_dir / 'README.md', 'rb'),
        )
    )



#!/usr/bin/env python3
#
# Tools for building OpenSlide and its dependencies
#
# Copyright (c) 2023 Benjamin Gilbert
# All rights reserved.
#
# This script is free software: you can redistribute it and/or modify it
# under the terms of the GNU Lesser General Public License, version 2.1,
# as published by the Free Software Foundation.
#
# This script is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
# for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this script. If not, see <http://www.gnu.org/licenses/>.
#

from __future__ import annotations

from pathlib import Path
import shutil

from common.argparse import TypedArgs
from common.software import Project


class Args(TypedArgs):
    dir: Path


args = Args('write-licenses', description='Write licenses directory.')
args.add_arg('dir', type=Path, help='output directory')
args.parse()

if args.dir.exists():
    shutil.rmtree(args.dir)
for proj in Project.get_enabled():
    proj.write_license_files(args.dir)



#!/usr/bin/env python3
#
# generate a test program for the API
#

import xml.etree.ElementTree as etree
import os
import re
import sys

import xmlmod

# Globals

dtors = {
    'htmlDoc *': 'xmlFreeDoc',
    'htmlParserCtxt *': 'htmlFreeParserCtxt',
    'xmlAutomata *': 'xmlFreeAutomata',
    'xmlBuffer *': 'xmlBufferFree',
    'xmlCatalog *': 'xmlFreeCatalog',
    'xmlChar *': 'xmlFree',
    'xmlDOMWrapCtxt *': 'xmlDOMWrapFreeCtxt',
    'xmlDict *': 'xmlDictFree',
    'xmlDoc *': 'xmlFreeDoc',
    'xmlDtd *': 'xmlFreeDtd',
    'xmlEntitiesTable *': 'xmlFreeEntitiesTable',
    'xmlElementContent *': 'xmlFreeElementContent',
    'xmlEnumeration *': 'xmlFreeEnumeration',
    'xmlList *': 'xmlListDelete',
    'xmlModule *': 'xmlModuleFree',
    'xmlMutex *': 'xmlFreeMutex',
    'xmlNode *': 'xmlFreeNode',
    'xmlNodeSet *': 'xmlXPathFreeNodeSet',
    'xmlNs *': 'xmlFreeNs',
    'xmlOutputBuffer *': 'xmlOutputBufferClose',
    'xmlParserCtxt *': 'xmlFreeParserCtxt',
    'xmlParserInputBuffer *': 'xmlFreeParserInputBuffer',
    'xmlParserInput *': 'xmlFreeInputStream',
    'xmlRMutex *': 'xmlFreeRMutex',
    'xmlRelaxNGValidCtxt *': 'xmlRelaxNGFreeValidCtxt',
    'xmlSaveCtxt *': 'xmlSaveClose',
    'xmlSchemaFacet *': 'xmlSchemaFreeFacet',
    'xmlSchemaVal *': 'xmlSchemaFreeValue',
    'xmlSchemaValidCtxt *': 'xmlSchemaFreeValidCtxt',
    'xmlTextWriter *': 'xmlFreeTextWriter',
    'xmlURI *': 'xmlFreeURI',
    'xmlValidCtxt *': 'xmlFreeValidCtxt',
    'xmlXPathContext *': 'xmlXPathFreeContext',
    'xmlXPathParserContext *': 'xmlXPathFreeParserContext',
    'xmlXPathObject *': 'xmlXPathFreeObject',
}

blockList = {
    # init/cleanup
    'xmlCleanupParser': True,
    'xmlInitParser': True,

    # arg must be non-NULL
    'xmlMemStrdupLoc': True,
    'xmlMemoryStrdup': True,

    # Returns void pointer which must be freed
    'xmlMallocAtomicLoc': True,
    'xmlMallocLoc': True,
    'xmlMemMalloc': True,
    'xmlMemRealloc': True,
    'xmlReallocLoc': True,

    # Would reset the error handler
    'xmlSetStructuredErrorFunc': True,

    # Prints errors
    'xmlCatalogGetPublic': True,
    'xmlCatalogGetSystem': True,
    'xmlDebugDumpDTD': True,
    'xmlDebugDumpDocument': True,
    'xmlDebugDumpNode': True,
    'xmlDebugDumpString': True,
    'xmlParserError': True,
    'xmlParserWarning': True,
    'xmlParserValidityError': True,
    'xmlParserValidityWarning': True,

    # Internal parser unctions, ctxt must be non-NULL
    'xmlParseAttribute': True,
    'xmlParseAttributeListDecl': True,
    'xmlParseAttributeType': True,
    'xmlParseCDSect': True,
    'xmlParseCharData': True,
    'xmlParseCharRef': True,
    'xmlParseComment': True,
    'xmlParseDefaultDecl': True,
    'xmlParseDocTypeDecl': True,
    'xmlParseEndTag': True,
    'xmlParseElement': True,
    'xmlParseElementChildrenContentDecl': True,
    'xmlParseElementContentDecl': True,
    'xmlParseElementDecl': True,
    'xmlParseElementMixedContentDecl': True,
    'xmlParseEncName': True,
    'xmlParseEncodingDecl': True,
    'xmlParseEntityDecl': True,
    'xmlParseEntityValue': True,
    'xmlParseEnumeratedType': True,
    'xmlParseEnumerationType': True,
    'xmlParseExternalID': True,
    'xmlParseExternalSubset': True,
    'xmlParseMarkupDecl': True,
    'xmlParseMisc': True,
    'xmlParseName': True,
    'xmlParseNmtoken': True,
    'xmlParseNotationDecl': True,
    'xmlParseNotationType': True,
    'xmlParsePEReference': True,
    'xmlParsePI': True,
    'xmlParsePITarget': True,
    'xmlParsePubidLiteral': True,
    'xmlParseReference': True,
    'xmlParseSDDecl': True,
    'xmlParseStartTag': True,
    'xmlParseSystemLiteral': True,
    'xmlParseTextDecl': True,
    'xmlParseVersionInfo': True,
    'xmlParseVersionNum': True,
    'xmlParseXMLDecl': True,
    'xmlParserHandlePEReference': True,
    'xmlSkipBlankChars': True,

    # reads from stdin
    'htmlReadFd': True,
    'xmlReadFd': True,
    'xmlReaderForFd': True,
}

# Parse document

if len(sys.argv) > 1:
    buildDir = sys.argv[1]
else:
    buildDir = '.'

xmlDocDir = buildDir + '/doc/xml'

filenames = {}
functions = {}

for file in os.listdir(xmlDocDir):
    if not file.endswith('_8h.xml'):
        continue

    doc = etree.parse(xmlDocDir + '/' + file)

    compound = doc.find('compounddef')
    module = compound.find('compoundname').text
    if not module.endswith('.h'):
        continue
    module = module[:-2]

    for section in compound.findall('sectiondef'):
        if section.get('kind') != 'func':
            continue

        for func in section.findall('memberdef'):
            name = func.find('name').text
            if name in blockList:
                continue

            module1, module2 = xmlmod.findModules(module, name)

            cargs = []
            skip = False
            for arg in func.findall('param'):
                atype = etree.tostring(arg.find('type'),
                    method='text', encoding='unicode').rstrip()
                if atype == 'void':
                    continue
                if atype == 'va_list':
                    skip = True
                    break
                if re.search(r'(Ptr|\*)$', atype):
                    cargs.append('NULL')
                else:
                    cargs.append('0')

            if skip:
                continue

            mfunc = functions.get(module1)
            if mfunc is None:
                mfunc = {}
                functions[module1] = mfunc

            mmfunc = mfunc.get(module2)
            if mmfunc is None:
                mmfunc = {}
                mfunc[module2] = mmfunc

            code = f'{name}({', '.join(cargs)})'

            rtype = etree.tostring(func.find('type'),
                method='text', encoding='unicode').rstrip()
            dtor = dtors.get(rtype)
            if dtor is not None:
                code = f'{dtor}({code})'
            elif rtype == 'xmlHashTable *':
                code = f'xmlHashFree({code}, NULL)'

            mmfunc[name] = f'    {code};\n'

# Write output

test = open('testapi.c', 'w')

test.write("""/*
 * testapi.c: libxml2 API tester program.
 *
 * Automatically generated by gentest.py
 *
 * See Copyright for the status of this software.
 */

/* Disable deprecation warnings */
#define XML_DEPRECATED

#include "libxml.h"
#include <libxml/HTMLparser.h>
#include <libxml/HTMLtree.h>
#include <libxml/c14n.h>
#include <libxml/catalog.h>
#include <libxml/debugXML.h>
#include <libxml/parserInternals.h>
#include <libxml/pattern.h>
#include <libxml/relaxng.h>
#include <libxml/schematron.h>
#include <libxml/uri.h>
#include <libxml/xinclude.h>
#include <libxml/xlink.h>
#include <libxml/xmlmodule.h>
#include <libxml/xmlreader.h>
#include <libxml/xmlsave.h>
#include <libxml/xmlschemas.h>
#include <libxml/xmlschemastypes.h>
#include <libxml/xmlwriter.h>
#include <libxml/xpathInternals.h>
#include <libxml/xpointer.h>

static void
ignoreError(void *userData ATTRIBUTE_UNUSED,
            const xmlError *error ATTRIBUTE_UNUSED) {
}

int
main(int argc ATTRIBUTE_UNUSED, char **argv ATTRIBUTE_UNUSED) {
    xmlInitParser();
    xmlSetStructuredErrorFunc(NULL, ignoreError);

""")

for module1 in sorted(functions.keys()):
    mfunc = functions[module1]

    if module1 != '':
        test.write(f'#ifdef LIBXML_{module1}_ENABLED\n')

    for module2 in sorted(mfunc.keys()):
        mmfunc = mfunc[module2]

        if module2 != '':
            test.write(f'#ifdef LIBXML_{module2}_ENABLED\n')

        for name in sorted(mmfunc.keys()):
            test.write(mmfunc[name])

        if module2 != '':
            test.write(f'#endif /* LIBXML_{module2}_ENABLED */\n')

    if module1 != '':
        test.write(f'#endif /* LIBXML_{module1}_ENABLED */\n')

    test.write('\n')

test.write("""    xmlCleanupParser();
    return 0;
}
""")



#!/usr/bin/env python3

import os
import sys
import subprocess

if 'DESTDIR' not in os.environ:
    bindir = sys.argv[1]
    libdir = sys.argv[2]
    binary_version = sys.argv[3]

    query_loaders = os.path.join(bindir, "gdk-pixbuf-query-loaders")
    loaders_dir = os.path.join(libdir, "gdk-pixbuf-2.0", binary_version)
    loaders_cache = os.path.join(loaders_dir, "loaders.cache")

    os.makedirs(loaders_dir, exist_ok=True)

    cmd = [query_loaders]
    with subprocess.Popen(cmd, stdout=subprocess.PIPE, close_fds=True) as p:
        data = p.stdout.read()

    with open(loaders_cache, "wb") as f:
        f.write(data)
else:
    print("*** Warning: loaders.cache not built because DESTDIR is set")
    print("***          You will need to manually call gdk-pixbuf-query-loaders")



#!/usr/bin/env python3
#
# Copyright © 2019 Endless Mobile, Inc.
#
# SPDX-License-Identifier: LGPL-2.1-or-later
#
# Original author: Philip Withnall

"""
Checks that a merge request doesn’t add any instances of the string ‘todo’
(in uppercase), or similar keywords. It may remove instances of that keyword,
or move them around, according to the logic of `git log -S`.
"""

import argparse
import re
import subprocess
import sys


# We have to specify these keywords obscurely to avoid the script matching
# itself. The keyword ‘fixme’ (in upper case) is explicitly allowed because
# that’s conventionally used as a way of marking a workaround which needs to
# be merged for now, but is to be grepped for and reverted or reworked later.
BANNED_KEYWORDS = ["TO" + "DO", "X" + "XX", "W" + "IP"]


def main():
    parser = argparse.ArgumentParser(
        description="Check a range of commits to ensure they don’t contain "
        "banned keywords."
    )
    parser.add_argument("commits", help="SHA to diff from, or range of commits to diff")
    args = parser.parse_args()

    banned_words_seen = set()
    seen_in_log = False
    seen_in_diff = False

    # Check the log messages for banned words.
    log_process = subprocess.run(
        ["git", "log", "--no-color", args.commits + "..HEAD"],
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
        encoding="utf-8",
        check=True,
    )
    log_lines = log_process.stdout.strip().split("\n")

    for line in log_lines:
        for keyword in BANNED_KEYWORDS:
            if re.search(r"(^|\W+){}(\W+|$)".format(keyword), line):
                banned_words_seen.add(keyword)
                seen_in_log = True

    # Check the diff for banned words.
    diff_process = subprocess.run(
        ["git", "diff", "-U0", "--no-color", args.commits],
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
        encoding="utf-8",
        check=True,
    )
    diff_lines = diff_process.stdout.strip().split("\n")

    for line in diff_lines:
        if not line.startswith("+ "):
            continue

        for keyword in BANNED_KEYWORDS:
            if re.search(r"(^|\W+){}(\W+|$)".format(keyword), line):
                banned_words_seen.add(keyword)
                seen_in_diff = True

    if banned_words_seen:
        if seen_in_log and seen_in_diff:
            where = "commit message and diff"
        elif seen_in_log:
            where = "commit message"
        elif seen_in_diff:
            where = "commit diff"

        print(
            "Saw banned keywords in a {}: {}. "
            "This indicates the branch is a work in progress and should not "
            "be merged in its current "
            "form.".format(where, ", ".join(banned_words_seen))
        )
        sys.exit(1)


if __name__ == "__main__":
    main()



#!/usr/bin/python3

import argparse
import os
from pathlib import Path
import re

parser = argparse.ArgumentParser()
parser.add_argument('type', choices=['zlib', 'zlib-ng'])
args = parser.parse_args()

if args.type == 'zlib':
    regex = re.compile('#define ZLIB_VERSION "([0-9.]+)\\.zlib-ng"$')
else:
    regex = re.compile('#define ZLIBNG_VERSION "([0-9.]+)"$')

zlib_h_in = (
    Path(os.environ['MESON_SOURCE_ROOT']) /
    os.environ['MESON_SUBDIR'] /
    'zlib.h.in'
)
for line in zlib_h_in.open():
    match = regex.match(line)
    if match:
        print(match.group(1))
        break
else:
    raise Exception('Specified version not found in ' + zlib_h_in.as_posix())



#!/usr/bin/env python3

import os
import sys
import re
import subprocess

# Put output files in the builddir
os.chdir(os.environ['MESON_BUILD_ROOT'])

infile1 = 'unwind-section-conftest.s'
outfile1 = 'unwind-section-conftest.o'
infile2 = 'unwind-section-conftest.c'

with open(infile1, 'w') as f:
    f.write('''
        .text
        .globl foo
        foo:
        jmp bar
        .section .eh_frame,"a",@unwind
        bar:
    ''')

with open(infile2, 'w') as f:
    f.write('''
        extern void foo();
        int main(){foo();}
    ''')

args = sys.argv[1:]
subprocess.check_call(args + ['-Wa,--fatal-warnings', '-c', infile1, '-o', outfile1],
                      stdout=subprocess.PIPE, universal_newlines=True)
subprocess.check_call(args + [infile2, outfile1], stdout=subprocess.PIPE, universal_newlines=True)

sys.exit(0)



#!/usr/bin/env python3

import json
import sys
from dataclasses import dataclass

# The basic idea is to find named character references using binary
# search. Since entity strings may not have a terminator, this doesn't
# work if one entity string is a prefix of another. In this case,
# we branch to a subtable after matching the prefix.
#
# We create separate initial tables based on the first character
# of the entity name.
#
# The following tables are generated:
#
# htmlEntAlpha:   start and end of initial tables, indexing into
#                 htmlEntValues
# htmlEntValues:  concatenation of all table values, which index into
#                 htmlEntStrings
# htmlEntStrings: variable sized records containing entity name,
#                 replacement and optionally the position of a
#                 subtable

try:
    with open('entities.json') as json_data:
        ents = json.load(json_data)
except FileNotFoundError:
    print('entities.json not found, try curl -LJO',
          'https://html.spec.whatwg.org/entities.json')
    sys.exit(1)

def to_cchars(s):
    r = []

    for c in s.encode():
        if c >= 0x20 and c <= 0x7E and c != ord("'") and c != ord('\\'):
            v = f"'{chr(c)}'"
        else:
            v = c
        r += [ v ]

    return r

@dataclass
class PrefixStackEntry:
    prefix: str
    table_id: int

@dataclass
class AlphaFixup:
    table_id: int
    char: int

@dataclass
class StringFixup:
    table_id: int
    string_index: int
    super_table_id: int
    super_offset: int

# Remove entity strings without trailing semicolon
keys = (key for key in ents.keys() if key.endswith(';'))

# Sort entity strings
keys = sorted(keys, key=lambda k: k[1:-1])

strings = []
tables = []
prefix_stack = []
alpha_fixups = []
string_fixups = []
for i in range(64):
    tables.append([])

for i, key in enumerate(keys):
    name = key[1:-1]

    next_name = None
    if i + 1 < len(keys):
        next_name = keys[i+1][1:-1]

    while prefix_stack and not name.startswith(prefix_stack[-1].prefix):
        prefix_stack.pop()

    # First character is initial prefix
    if not prefix_stack:
        table_id = len(tables)
        tables.append([])

        prefix_stack.append(PrefixStackEntry(name[0], table_id))
        alpha_fixups.append(AlphaFixup(table_id, ord(name[0]) % 64))

    string_index = len(strings)
    table = tables[prefix_stack[-1].table_id]
    table_index = len(table)
    table.append(string_index)

    name_offset = len(prefix_stack[-1].prefix)
    name_chars = to_cchars(name[name_offset:])
    repl_chars = to_cchars(ents[key]['characters'])
    semicolon_flag = 0
    if key[:-1] in ents:
        semicolon_flag = 0x80

    if next_name and next_name.startswith(name):
        # Create subtable

        strings += [
            len(name_chars) | semicolon_flag | 0x40, *name_chars,
            0, 0, # subtable position, to be fixed up
            len(repl_chars), *repl_chars,
        ]

        table_id = len(tables)
        tables.append([])

        fixup_index = string_index + 1 + len(name_chars)
        string_fixups.append(StringFixup(
            table_id, fixup_index, prefix_stack[-1].table_id, table_index,
        ))

        prefix_stack.append(PrefixStackEntry(name, table_id))
    else:
        strings += [
            len(name_chars) | semicolon_flag, *name_chars,
            len(repl_chars), *repl_chars,
        ]

# Concat tables and record ranges
ranges = [ 0 ]
values = []
for table in tables:
    values += table
    ranges.append(len(values))

# Create alpha table
alpha = [ 0 ] * (59 * 3)
for fixup in alpha_fixups:
    table_id, c = fixup.table_id, fixup.char
    start = ranges[table_id]
    end = ranges[table_id+1]
    alpha[c*3:c*3+3] = [ start & 0xFF, start >> 8, end - start ]

# Fix up subtable positions
for fixup in string_fixups:
    table_id, i = fixup.table_id, fixup.string_index
    start = ranges[table_id]
    end = ranges[table_id+1]
    super_index = ranges[fixup.super_table_id] + fixup.super_offset
    strings[i:i+2] = [ start - super_index, end - start ]

# Print tables

def gen_table(ctype, cname, values, fmt, elems_per_line):
    count = len(values)
    r = ''

    for i in range(count):
        if i != 0: r += ','
        if i % elems_per_line == 0: r += '\n    '
        else: r += ' '
        r += fmt % values[i]

    return f'static const {ctype} {cname}[{count}] = {{{r}\n}};\n\n'

with open('codegen/html5ent.inc', 'w') as out:
    out.write(gen_table('unsigned char', 'htmlEntAlpha', alpha, '%3d', 15))
    out.write(gen_table('unsigned short', 'htmlEntValues', values, '%5d', 10))
    out.write(gen_table('unsigned char', 'htmlEntStrings', strings, '%3s', 15))



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Copyright © 2025 Marco Trevisan <mail@3v1n0.net>
#
# SPDX-License-Identifier: LGPL-2.1-or-later
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
# MA  02110-1301  USA

"""Integration tests for gi-compile-repository."""

import os
import unittest

import glibconfig
import taptestrunner
import testprogramrunner


class TestGICompileRepositoryBase(testprogramrunner.TestProgramRunner):
    """Integration test base class for checking gi-compile-repository behavior"""

    PROGRAM_NAME = "gi-compile-repository"
    PROGRAM_TYPE = testprogramrunner.ProgramType.NATIVE

    @classmethod
    def setUpClass(cls):
        super().setUpClass()

        # list of pathlib.Path
        cls._gir_paths = glibconfig.GIR_XML_SEARCH_PATHS
        print(f"gir path set to {cls._gir_paths}")


class TestGICompileRepository(TestGICompileRepositoryBase):
    """Integration test for checking gi-compile-repository behavior"""

    def test_open_failure(self):
        gir_path = "this-is/not/a-file.gir"
        result = self.runTestProgram(
            [gir_path, "--output", os.path.join(self.tmpdir.name, "invalid.typelib")],
            should_fail=True,
        )

        self.assertEqual(result.info.returncode, 1)
        self.assertIn(f"Error parsing file ‘{gir_path}’", result.err)


class TestGICompileRepositoryForGLib(TestGICompileRepositoryBase):
    GIR_NAME = "GLib-2.0"

    def runTestProgram(self, *args, **kwargs):
        for d in self._gir_paths:
            gir_file = d / f"{self.GIR_NAME}.gir"

            if gir_file.exists():
                break
        else:
            self.fail(f"Could not find {self.GIR_NAME}.gir in {self._gir_paths}")

        argv = [str(gir_file)]
        argv.extend(*args)

        if self.GIR_NAME != "GLib-2.0":
            for d in self._gir_paths:
                argv.extend(["--includedir", str(d)])

        return super().runTestProgram(argv, **kwargs)

    def test_write_failure(self):
        typelib_path = "this-is/not/a-good-output/invalid.typelib"
        result = self.runTestProgram(
            ["--output", typelib_path],
            should_fail=True,
        )

        self.assertEqual(result.info.returncode, 1)
        self.assertIn(f"Failed to open ‘{typelib_path}.tmp’", result.err)

    def test_compile(self):
        typelib_name = os.path.splitext(self.GIR_NAME)[0]
        typelib_path = os.path.join(self.tmpdir.name, f"{typelib_name}.typelib")
        argv = ["--output", typelib_path]

        result = self.runTestProgram(argv)

        self.assertFalse(result.out)
        self.assertFalse(result.err)
        self.assertTrue(os.path.exists(typelib_path))


class TestGICompileRepositoryForGObject(TestGICompileRepositoryForGLib):
    GIR_NAME = "GObject-2.0"


class TestGICompileRepositoryForGio(TestGICompileRepositoryForGLib):
    GIR_NAME = "Gio-2.0"


if __name__ == "__main__":
    unittest.main(testRunner=taptestrunner.TAPTestRunner())



#
# Tools for building OpenSlide and its dependencies
#
# Copyright (c) 2023-2025 Benjamin Gilbert
# All rights reserved.
#
# This script is free software: you can redistribute it and/or modify it
# under the terms of the GNU Lesser General Public License, version 2.1,
# as published by the Free Software Foundation.
#
# This script is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
# for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this script. If not, see <http://www.gnu.org/licenses/>.
#

from __future__ import annotations

from email.message import Message
from email.policy import Compat32
from pathlib import PurePath
import re
import tomllib

from .meson import meson_introspect, meson_source_root
from .software import Project, get_spdx


def pyproject_fill_template(tmpl: str) -> str:
    version: str = meson_introspect('projectinfo')['version']
    projs = Project.get_enabled()
    license_relpaths = [PurePath('COPYING.LESSER')] + [
        PurePath('licenses') / relpath
        for proj in projs
        for relpath in proj.license_relpaths
    ]
    for relpath in license_relpaths:
        # we don't do proper toml assembly
        if '"' in relpath.as_posix():
            raise ValueError(f'Invalid license file path: {relpath}')
    return (
        tmpl.replace('@version@', version)
        .replace('@spdx@', get_spdx(projs))
        .replace(
            '"@license-files@"',
            '["'
            + '", "'.join(
                sorted(
                    (relpath.as_posix() for relpath in license_relpaths),
                    key=lambda s: s.lower(),
                )
            )
            + '"]',
        )
    )


def pyproject_to_message(pyproject: str) -> Message:
    meta = tomllib.loads(pyproject)
    out = Message(policy=Compat32(max_line_length=None))
    out['Metadata-Version'] = '2.4'
    for k, v in meta['project'].items():
        k = k.lower()
        if k == 'name':
            out['Name'] = v
        elif k == 'version':
            out['Version'] = v
        elif k == 'description':
            out['Summary'] = v
        elif k == 'readme':
            out.set_payload((meson_source_root() / v).read_bytes())
            if v.lower().endswith('.md'):
                out['Description-Content-Type'] = 'text/markdown'
        elif k == 'keywords':
            out['Keywords'] = ','.join(v)
        elif k == 'urls':
            for kk, vv in v.items():
                kk = kk.lower()
                if kk == 'homepage':
                    out['Home-page'] = vv
                elif kk == 'release notes':
                    out['Project-URL'] = f'Release notes, {vv}'
                elif kk == 'repository':
                    out['Project-URL'] = f'Repository, {vv}'
                else:
                    raise Exception(f'Unknown URL type: {kk}')
        elif k == 'authors':
            for item in v:
                out['Author-Email'] = f'{item["name"]} <{item["email"]}>'
        elif k == 'maintainers':
            for item in v:
                out['Maintainer-Email'] = f'{item["name"]} <{item["email"]}>'
        elif k == 'license':
            out['License-Expression'] = v
        elif k == 'license-files':
            for vv in v:
                if re.search(r'\*|\?|\[|\.\.', vv):
                    raise ValueError(
                        f'Glob or path traversal in license file path: {vv}'
                    )
                out['License-File'] = vv
        elif k == 'classifiers':
            for vv in v:
                out['Classifier'] = vv
        elif k == 'requires-python':
            out['Requires-Python'] = v
        else:
            raise Exception(f'Unknown field: {k}')
    return out



# Symbol to module mapping
#
# This relies on a few tables and some regexes.

import re

moduleMap = {
    'HTMLtree': 'HTML',
    'HTMLparser': 'HTML',
    'c14n': 'C14N',
    'catalog': 'CATALOG',
    'debugXML': 'DEBUG',
    'nanohttp': 'HTTP',
    'pattern': 'PATTERN',
    'relaxng': 'RELAXNG',
    'schemasInternals': 'SCHEMAS',
    'schematron': 'SCHEMATRON',
    'xinclude': 'XINCLUDE',
    'xlink': 'XPTR',
    'xmlautomata': 'REGEXP',
    'xmlmodule': 'MODULES',
    'xmlreader': 'READER',
    'xmlregexp': 'REGEXP',
    'xmlsave': 'OUTPUT',
    'xmlschemas': 'SCHEMAS',
    'xmlschemastypes': 'SCHEMAS',
    'xmlwriter': 'WRITER',
    'xpath': 'XPATH',
    'xpathInternals': 'XPATH',
    'xpointer': 'XPTR',
}

symbolMap1 = {
    # not VALID
    'xmlValidateNCName': '',
    'xmlValidateNMToken': '',
    'xmlValidateName': '',
    'xmlValidateQName': '',

    'htmlDefaultSAXHandlerInit': 'HTML',
    'xmlSAX2InitHtmlDefaultSAXHandler': 'HTML',

    'xmlRegisterHTTPPostCallbacks': 'HTTP',

    '__xmlOutputBufferCreateFilename': 'OUTPUT',
    'xmlAttrSerializeTxtContent': 'OUTPUT',
    'xmlUTF8ToIsolat1': 'OUTPUT',
    'xmlSprintfElementContent': 'OUTPUT',

    'xmlCreatePushParserCtxt': 'PUSH',
    'xmlParseChunk': 'PUSH',

    'xmlParseBalancedChunkMemory': 'SAX1',
    'xmlParseBalancedChunkMemoryRecover': 'SAX1',
    'xmlParseDoc': 'SAX1',
    'xmlParseEntity': 'SAX1',
    'xmlParseExternalEntity': 'SAX1',
    'xmlParseFile': 'SAX1',
    'xmlParseMemory': 'SAX1',
    'xmlSAXDefaultVersion': 'SAX1',
    'xmlSetupParserForBuffer': 'SAX1',

    'xmlCtxtGetValidCtxt': 'VALID',
    'xmlFreeValidCtxt': 'VALID',
    'xmlNewValidCtxt': 'VALID',

    'xmlCatalogConvert': 'SGML_CATALOG',
    'xmlConvertSGMLCatalog': 'SGML_CATALOG',
    'xmlLoadSGMLSuperCatalog': 'SGML_CATALOG',
}

symbolMap2 = {
    # not OUTPUT (should be fixed in xmlIO.h)
    'xmlOutputBufferCreateFilenameDefault': '',

    'xmlXPathDebugDumpCompExpr': 'DEBUG',
    'xmlXPathDebugDumpObject': 'DEBUG',
    'xmlSchemaDump': 'DEBUG',
    'xmlRelaxNGDump': 'DEBUG',

    'xmlACatalogDump': 'OUTPUT',
    'xmlCatalogDump': 'OUTPUT',
    'xmlIOHTTPOpenW': 'OUTPUT',
    'xmlNanoHTTPSave': 'OUTPUT',
    'xmlRegisterHTTPPostCallbacks': 'OUTPUT',
    'xmlRelaxNGDumpTree': 'OUTPUT',

    'xmlTextReaderPreservePattern': 'PATTERN',

    'htmlCreatePushParserCtxt': 'PUSH',
    'htmlParseChunk': 'PUSH',

    'xmlValidBuildContentModel': 'REGEXP',
    'xmlValidatePopElement': 'REGEXP',
    'xmlValidatePushCData': 'REGEXP',
    'xmlValidatePushElement': 'REGEXP',

    'xmlTextReaderRelaxNGSetSchema': 'RELAXNG',
    'xmlTextReaderRelaxNGValidate': 'RELAXNG',
    'xmlTextReaderRelaxNGValidateCtxt': 'RELAXNG',

    'xmlTextReaderSchemaValidate': 'SCHEMAS',
    'xmlTextReaderSchemaValidateCtxt': 'SCHEMAS',
    'xmlTextReaderSetSchema': 'SCHEMAS',

    'xmlTextReaderReadInnerXml': 'WRITER',
    'xmlTextReaderReadOuterXml': 'WRITER',
}

outputRegex = '|'.join((
    '^(html|xml(Buf)?)(Doc(Content|Format)?|Elem|Node)Dump',
    '^(html|xml)Save(Format)?File',
    '^xmlDump.*(Decl|Table)',
    '^xml(Alloc)?OutputBuffer',
    '^xml.*OutputCallbacks',
))

def findModules(filename, symbol):
    module1 = symbolMap1.get(symbol)

    if module1 is None:
        module1 = moduleMap.get(filename)

    if module1 is None:
        if re.search('^xml(Ctxt)?Valid|Parse(DTD|Dtd)', symbol):
            module1 = 'VALID'
        elif re.search('^xml(Recover|SAX(User)?Parse)', symbol):
            module1 = 'SAX1'
        elif re.search('^xmlIOHTTP', symbol):
            module1 = 'HTTP'

    module2 = symbolMap2.get(symbol)

    if module2 is None:
        if re.search(outputRegex, symbol):
            if module1 is None:
                module1 = 'OUTPUT'
            else:
                module2 = 'OUTPUT'

    if module1 is None:
        module1 = ''
    if module2 is None:
        module2 = ''

    return module1, module2



#!/usr/bin/env python3
#
# Copyright © 2021 Chun-wei Fan.
#
# SPDX-License-Identifier: LGPL-2.1-or-later
#
# Original author: Chun-wei Fan <fanc999@yahoo.com.tw>

"""
This script generates a Windows manifest file and optionally a resource file to
determine whether a specified program requires UAC elevation
"""

import os
import argparse

DOMAIN_NAME = "gnome"


def main():
    parser = argparse.ArgumentParser(description=__doc__)
    parser.add_argument(
        "-p", "--package", required=True, help="package name of the executable"
    )
    parser.add_argument("-n", "--name", required=True, help="name of executable")
    parser.add_argument(
        "--pkg-version", required=True, dest="version", help="version of package"
    )
    parser.add_argument(
        "--require-admin",
        action="store_true",
        dest="admin",
        default=False,
        help="require admin access to application",
    )
    parser.add_argument(
        "--input-resource-file",
        dest="resource",
        default=None,
        help="existing .rc file to embed UAC manifest (do not generate a new .rc file), must have included windows.h in it",
    )
    parser.add_argument(
        "--output-dir",
        dest="outdir",
        default=None,
        help="directory to output resulting files",
    )
    args = parser.parse_args()

    if args.resource is not None:
        if not os.path.isfile(args.resource):
            raise FileNotFoundError(
                "Specified resource file '%s' does not exist" % args.resource
            )

    generate_manifest(args.package, args.name, args.version, args.admin, args.outdir)
    write_rc_file(args.name, args.resource, args.outdir)


def generate_manifest(package, name, version, admin, outdir):
    if version.count(".") == 0:
        manifest_package_version = version + ".0.0.0"
    elif version.count(".") == 1:
        manifest_package_version = version + ".0.0"
    elif version.count(".") == 2:
        manifest_package_version = version + ".0"
    elif version.count(".") == 3:
        manifest_package_version = version
    else:
        parts = version.split(".")
        manifest_package_version = ""
        for x in (0, 1, 2, 3):
            if x == 0:
                manifest_package_version += parts[x]
            else:
                manifest_package_version += "." + parts[x]

    if outdir is not None:
        output_file_base_name = os.path.join(outdir, name)
    else:
        output_file_base_name = name

    outfile = open(output_file_base_name + ".exe.manifest", "w+")
    outfile.write("<?xml version='1.0' encoding='UTF-8' standalone='yes'?>\n")
    outfile.write(
        "<assembly xmlns='urn:schemas-microsoft-com:asm.v1' manifestVersion='1.0'>\n"
    )
    outfile.write("  <assemblyIdentity version='%s'\n" % manifest_package_version)
    outfile.write("    processorArchitecture='*'\n")
    outfile.write("    name='%s.%s.%s.exe'\n" % (DOMAIN_NAME, package, name))
    outfile.write("    type='win32' />\n")
    outfile.write("  <trustInfo xmlns='urn:schemas-microsoft-com:asm.v3'>\n")
    outfile.write("    <security>\n")
    outfile.write("      <requestedPrivileges>\n")
    outfile.write("        <requestedExecutionLevel\n")

    if admin:
        outfile.write("          level='requireAdministrator'\n")
    else:
        outfile.write("          level='asInvoker'\n")

    outfile.write("          uiAccess='false' />\n")
    outfile.write("      </requestedPrivileges>\n")
    outfile.write("    </security>\n")
    outfile.write("  </trustInfo>\n")
    outfile.write("</assembly>\n")
    outfile.close()


def write_rc_file(name, resource, outdir):
    if outdir is not None:
        output_file_base_name = os.path.join(outdir, name)
    else:
        output_file_base_name = name

    if resource is None:
        outfile = open(output_file_base_name + ".rc", "w+")
        outfile.write("#define WIN32_LEAN_AND_MEAN\n")
        outfile.write("#include <windows.h>\n")
    else:
        if resource != output_file_base_name + ".rc":
            outfile = open(output_file_base_name + ".rc", "w+")
        else:
            outfile = open(output_file_base_name + ".final.rc", "w+")
        srcfile = open(resource, "r")
        outfile.write(srcfile.read())
        srcfile.close()

    outfile.write("\n")
    outfile.write(
        'CREATEPROCESS_MANIFEST_RESOURCE_ID RT_MANIFEST "%s.exe.manifest"' % name
    )
    outfile.close()


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
#
# Tools for building OpenSlide and its dependencies
#
# Copyright (c) 2011-2015 Carnegie Mellon University
# Copyright (c) 2022-2023 Benjamin Gilbert
# All rights reserved.
#
# This script is free software: you can redistribute it and/or modify it
# under the terms of the GNU Lesser General Public License, version 2.1,
# as published by the Free Software Foundation.
#
# This script is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
# for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this script. If not, see <http://www.gnu.org/licenses/>.
#

from __future__ import annotations

import os
from pathlib import Path
import re
import subprocess

from common.argparse import TypedArgs
from common.meson import meson_host


def library_symbols(file: Path) -> list[str]:
    if host == 'linux':
        out = subprocess.check_output(
            [os.environ['OBJDUMP'], '-T', file]
        ).decode()
        return [line.split()[6] for line in out.split('\n') if '.text' in line]
    elif host == 'darwin':
        out = subprocess.check_output(
            [os.environ['DYLD_INFO'], '-exports', file]
        ).decode()
        return [
            line.split()[1].lstrip('_')
            for line in out.split('\n')
            if ' 0x' in line
        ]
    elif host == 'windows':
        out = subprocess.check_output(
            [os.environ['OBJDUMP'], '-p', file]
        ).decode()
        active = False
        syms: list[str] = []
        for line in out.split('\n'):
            if active:
                if not line.strip():
                    return syms
                sym = line.split()[-1]
                if sym != 'Name':
                    syms.append(sym)
            elif 'Ordinal/Name Pointer' in line:
                active = True
        raise Exception("Couldn't parse objdump output")
    else:
        raise Exception(f'Unknown host: {host}')


class Args(TypedArgs):
    file: Path
    output: Path
    debuginfo: Path


args = Args(
    'postprocess-binary', description='Mangle shared library or executable.'
)
args.add_arg('-o', '--output', type=Path, required=True, help='output file')
args.add_arg(
    '-d', '--debuginfo', type=Path, required=True, help='output debug symbols'
)
args.add_arg('file', type=Path, help='input file')
args.parse()
host = meson_host()

# split debuginfo
if host == 'darwin':
    subprocess.check_call(
        [os.environ['DSYMUTIL'], '-o', args.debuginfo, args.file]
    )
    subprocess.check_call(
        [os.environ['STRIP'], '-u', '-r', '-o', args.output, args.file]
    )
else:
    objcopy = os.environ['OBJCOPY']
    subprocess.check_call(
        [objcopy, '--only-keep-debug', args.file, args.debuginfo]
    )
    os.chmod(args.debuginfo, 0o644)
    # debuglink without a directory path enables search semantics
    assert args.debuginfo.parent == args.output.parent
    subprocess.check_call(
        [
            objcopy,
            '-S',
            f'--add-gnu-debuglink={args.debuginfo.name}',
            args.file.absolute(),
            args.output.absolute(),
        ],
        cwd=args.debuginfo.parent,
    )

# check for extra symbol exports
if re.search('\\.(dll|dylib|so[.0-9]*)$', args.file.name):
    syms = library_symbols(args.file)
    if not syms:
        raise Exception(f"Couldn't find exported symbols in {args.file}")
    syms = [
        # filter out acceptable symbols
        sym
        for sym in syms
        if not sym.startswith('openslide_')
    ]
    if syms:
        raise Exception(f'Unexpected exports in {args.file}: {syms}')

# update rpath
if host == 'linux' and not re.match('\\.so[.0-9]*$', args.file.name):
    subprocess.check_call(
        [os.environ['PATCHELF'], '--set-rpath', '$ORIGIN/../lib', args.output]
    )
elif host == 'darwin' and not args.file.name.endswith('.dylib'):
    out = subprocess.check_output(
        [os.environ['OTOOL'], '-l', args.output]
    ).decode()
    active = False
    for line in out.split('\n'):
        if 'cmd LC_RPATH' in line:
            active = True
        elif active:
            words = line.split()
            if words[0] == 'path':
                old_rpath = words[1]
                break
    else:
        raise Exception("Couldn't read LC_RPATH")
    subprocess.check_call(
        [
            os.environ['INSTALL_NAME_TOOL'],
            '-rpath',
            old_rpath,
            '@loader_path/../lib',
            args.output,
        ]
    )



# SPDX-FileCopyrightText: 2023 Guido Günther
# base on # codegen_rst.py (C) 2022 Emmanuele Bassi
#
# SPDX-License-Identifier: LGPL-2.1-or-later

import os
import re

import textwrap

# Disable line length warnings as wrapping the templates would be hard
# flake8: noqa: E501


class MdCodeGenerator:
    """Generates documentation in Markdown format."""

    def __init__(self, ifaces):
        self.ifaces = ifaces
        self._generate_expand_dicts()

    def _expand(self, s, expandParamsAndConstants):
        """Expands parameters and constant literals."""
        res = []
        for line in textwrap.dedent(s).split("\n"):
            line = line.rstrip()
            if line == "":
                res.append("")
                continue
            for key in self._expand_member_dict_keys:
                line = line.replace(key, self._expand_member_dict[key])
            for key in self._expand_iface_dict_keys:
                line = line.replace(key, self._expand_iface_dict[key])
            if expandParamsAndConstants:
                # replace @foo with `foo`
                line = re.sub(
                    "@[a-zA-Z0-9_]*",
                    lambda m: "`" + m.group(0)[1:] + "`",
                    line,
                )
                # replace e.g. %TRUE with ``TRUE``
                line = re.sub(
                    "%[a-zA-Z0-9_]*",
                    lambda m: "`" + m.group(0)[1:] + "`",
                    line,
                )
            res.append(line)
        return "\n".join(res)

    def _generate_expand_dicts(self):
        """Generates the dictionaries used to expand gtk-doc sigils."""
        self._expand_member_dict = {}
        self._expand_iface_dict = {}
        for i in self.ifaces:
            key = f"#{i.name}"
            value = f"`{i.name}`_"
            self._expand_iface_dict[key] = value

            for m in i.methods:
                key = "%s.%s()" % (i.name, m.name)
                value = f"`{i.name}.{m.name}`_"
                self._expand_member_dict[key] = value

            for s in i.signals:
                key = "#%s::%s" % (i.name, s.name)
                value = f"`{i.name}::{s.name}`_"
                self._expand_member_dict[key] = value

            for p in i.properties:
                key = "#%s:%s" % (i.name, p.name)
                value = f"`{i.name}:{p.name}`_"
                self._expand_member_dict[key] = value

        # Make sure to expand the keys in reverse order so e.g. #org.foo.Iface:MediaCompat
        # is evaluated before #org.foo.Iface:Media ...
        self._expand_member_dict_keys = sorted(
            self._expand_member_dict.keys(), reverse=True
        )
        self._expand_iface_dict_keys = sorted(
            self._expand_iface_dict.keys(), reverse=True
        )

    def _generate_header(self, iface):
        """Generates the header and preamble of the document."""
        header_len = len(iface.name)
        res = [
            f"Title: {iface.name} D-Bus Interface",
            f"Slug: {iface.name}",
            "",
            "# " + iface.name_without_prefix,
            "",
            "## Description",
            "",
            iface.doc_string_brief.strip(),
            "",
            self._expand(iface.doc_string, True),
            "",
        ]
        if iface.since:
            res += [
                f"Interface available since: {iface.since}.",
                "",
            ]
        if iface.deprecated:
            res += [
                "*Warning*: This interface is deprecated.",
                "",
            ]
        res += [""]
        return "\n".join(res)

    def _generate_section(self, title, name):
        """Generates a section with the given title."""
        res = [
            "### " + title,
            "",
        ]
        return "\n".join(res)

    def _generate_properties(self, iface):
        """Generates the properties section."""
        res = []
        for p in iface.properties:
            title = f"{iface.name}:{p.name}"
            if p.readable and p.writable:
                access = "readwrite"
            elif p.writable:
                access = "writable"
            else:
                access = "readable"
            res += [
                "### " + title,
                "",
                "```",
                f"    {p.name} {access} {p.signature}",
                "```",
                "",
                self._expand(p.doc_string, True),
                "",
            ]
            if p.since:
                res += [
                    f"Property available since: {p.since}.",
                    "",
                ]
            if p.deprecated:
                res += [
                    "*Warning*: This property is deprecated.",
                    "",
                ]
            res += [""]
        return "\n".join(res)

    def _generate_method_signature(self, method):
        """Generates the method signature as a code block."""
        res = [
            "```",
        ]
        n_in_args = len(method.in_args)
        n_out_args = len(method.out_args)
        if n_in_args == 0 and n_out_args == 0:
            res += [
                f"    {method.name} ()",
            ]
        else:
            res += [
                f"    {method.name} (",
            ]
            for idx, arg in enumerate(method.in_args):
                if idx == n_in_args - 1 and n_out_args == 0:
                    res += [
                        f"      IN {arg.name} {arg.signature}",
                    ]
                else:
                    res += [
                        f"      IN {arg.name} {arg.signature},",
                    ]
            for idx, arg in enumerate(method.out_args):
                if idx == n_out_args - 1:
                    res += [
                        f"      OUT {arg.name} {arg.signature}",
                    ]
                else:
                    res += [
                        f"      OUT {arg.name} {arg.signature},",
                    ]
            res += [
                "    )",
            ]
        res += ["```"]
        return "\n".join(res)

    def _generate_methods(self, iface):
        """Generates the methods section."""
        res = []
        for m in iface.methods:
            title = f"{iface.name}.{m.name}"
            res += [
                "### " + title,
                "",
                self._generate_method_signature(m),
                "",
                self._expand(m.doc_string, True),
                "",
            ]
            for a in m.in_args:
                arg_desc = self._expand(a.doc_string, True)
                res += [
                    f"* {a.name}: {arg_desc}",
                    "",
                ]
            res += [""]
            if m.since:
                res += [
                    f"Method available since: {m.since}.",
                    "",
                ]
            if m.deprecated:
                res += [
                    "*Warning*: This method is deprecated.",
                    "",
                ]
            res += [""]
        return "\n".join(res)

    def _generate_signal_signature(self, signal):
        """Generates the signal signature."""
        res = [
            "```",
        ]
        n_args = len(signal.args)
        if n_args == 0:
            res += [
                f"    {signal.name} ()",
            ]
        else:
            res += [
                f"    {signal.name} (",
            ]
            for idx, arg in enumerate(signal.args):
                if idx == n_args - 1:
                    res += [
                        f"      {arg.name} {arg.signature}",
                    ]
                else:
                    res += [
                        f"      {arg.name} {arg.signature},",
                    ]
            res += [
                "    )",
            ]
        res += ["```"]
        return "\n".join(res)

    def _generate_signals(self, iface):
        """Generates the signals section."""
        res = []
        for s in iface.signals:
            title = f"{iface.name}::{s.name}"
            res += [
                "### " + title,
                "",
                self._generate_signal_signature(s),
                "",
                self._expand(s.doc_string, True),
                "",
            ]
            for a in s.args:
                arg_desc = self._expand(a.doc_string, True)
                res += [
                    f"{a.name}",
                    f"  {arg_desc}",
                    "",
                ]
            res += [""]
            if s.since:
                res += [
                    f"Signal available since: {s.since}.",
                    "",
                ]
            if s.deprecated:
                res += [
                    "*Warning*: This signal is deprecated.",
                    "",
                ]
            res += [""]
        return "\n".join(res)

    def generate(self, md, outdir):
        """Generates the Markdown file for each interface."""
        for i in self.ifaces:
            with open(os.path.join(outdir, f"{md}-{i.name}.md"), "w") as outfile:
                outfile.write(self._generate_header(i))
                if len(i.properties) > 0:
                    outfile.write(self._generate_section("Properties", i.name))
                    outfile.write(self._generate_properties(i))
                if len(i.methods) > 0:
                    outfile.write(self._generate_section("Methods", i.name))
                    outfile.write(self._generate_methods(i))
                if len(i.signals) > 0:
                    outfile.write(self._generate_section("Signals", i.name))
                    outfile.write(self._generate_signals(i))



#!/usr/bin/env python3

import sys

with open(sys.argv[1], 'r') as f:
    for line in f.readlines():
        line = line.strip()
        if line and not line.startswith('#'):
            print(line)
            sys.exit(0)

sys.exit(1)
